'''
Created on June 20, 2018
@author: malemu
Description: Python script to run daily to retrieve and process and store data
                for Water point products
                the process takes in the following order:
                1) Get data
                2) sanitize input data from .txt files and build pgSQL
                    database queries
                3) calculate means
                4) execute queries on DB server.
******************************************************************************
Purpose:    This process gets rainfall(RFE) data for yesterday's date
            (self.run_date), perform various processes such as
            Unzip and untar, Derive today's; yesterdays and day before
            yesterday's date; Convert bil to grid, Subset to Africa,
            Project to LAEA (Lamberts Azimuthal Equal Area) Projection,
            Generate runoff grids from trmm rainfall (0.15*trmm),
            Zonalmean (TRMM, ET and Runoff girds with basins as zone grid),
            Band Math (depth calculations), Sample (pulls out depth values
            for each basin; output to text file),
            Append depth values to Text file

Algorithm:  A simple water balance approach is adopted.
                Wi = Wi-1 + PPTi - Runoffi - ETi
            water (W) is monitored using a water balance model,
            where PPT, Runoff and ET are daily measurements.

            The amount of water is fist converted to volume by multiplying
            the amount of water with Area of the basin.
            Then, the volume is converted to water depths by dividing with
            the area of the waterhole/lake or pond.

WORKSPACE:  The workspace for this program is
            G:\FEWSNET\Mali\Data\2016 It is predefined in the code.


INPUTS:     Rainfall from TRMM (1 degree grids),
            ET derived from GLDAS parameters (1 degree grids).
            To download data using FTP, a supporting file called download.txt
            is provided in the workspace which has all the DOS commands for
            ftp download.

            A predefined grid which has Africa's Lamberts
            Azimuthal Equal Area projection is required.
            Please copy it in the workspace before you run this code.

            A basins grid is required to perform
            zonalmean analysis of the rainfall,ET and runoff datasets.

            The basins coverage should be copied in the workspace prior to
            running this code.

            A grid which has point locations (single pixel grid) represnting
            each water hole is required to run extract the final depth values
            from each waterhole.

            Hence this grid file should be copied in the workspace prior to
            running this code.


OUTPUTS:    A simple text file which will have waterhole id in one column and
            waterhole water depth in another column. The text file will
            have waterhole depth information for 84 waterholes spread across
            the borders of Kenya and Ethiopia
            This can be used to plot graphs of daily waterhole water height
            variations for each waterhole. The output location is: "C:\LEWS"



USAGE:      The program can be invoked automatically by the system processess.
            There is no need to provide any parameters. The code calculates
            dates
            and correspondingly download the data and process them.
            The code is written in such away that it is very simple to use one.

@author:    Manohar Velpuri: First completed in December 2013
            Additional modifications made on September, 19th 2014
            Python version: Michael Alemu July,26 2018
******************************************************************************
Updated on:
    01/17/2019 - CHOLEN - Fixed Jdate and year variables to use the data date.
    05/02/2019 - CYOUNG - Added web_txt_distribution in formatOutput() to copy
                            required TXT files on web. Applied some fixes.
    06/27/2019 - ABALLU - Refactored the code and implemented some code
                            improvements
    07/11/2019 - ABALLU - Modified code to delete the arc grids which are two
                            days before run_date
    10/23/2019 - ABALLU - Modified code to delete the arc grids which are 90
                            days before run_date and made some minor code
                            improvements

'''
# Import standard modules
import calendar
import datetime as DT
import glob
import os
import re
import sys

# Import non-standard python modules.
import numpy as np
import pandas as pd
from dbfread import DBF

# Import ArcGIS python (arcpy) module
import arcpy

# Import bin modules
from bin import master_config as mconfig

# Import lib modules
from lib import check_input_args as chckargs
from lib import constants
from lib import dates
from lib import email_file
from lib import email_list
from lib import file_utilities
from lib import geo_transformation as gt
from lib import pg_sql_handler as cursor
from lib import print_msg
from lib import zips
from lib import temp_workspaces

# Check for arcpy extension and change overwrite settings
arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")


def checkInputArgs(config):
    """ Check input arguments """
    run_date = None
    run_processes = False
    steps = None
    today = DT.datetime.today()
    steps_comp = []
    proc_usage = ("\nUSAGE: <python> {} [Steps] <YYYY> <M> <D>"
                  "\n\twhere{}"
                  "\n\n\tYYYY is the 4-digit year "
                  "\n\tM is the the month of the year (1-12)"
                  "\n\tD is the the day of the month (e.g. 1-30 for Jan)"
                  "\n\n\tYYYY M D is an optional input, so the default input "
                  "information  is today's date").format(
        os.path.basename(sys.argv[0]),
        config.steps_desc)
    nargs_withoutpd = 2
    nargs_withpd = nargs_withoutpd + 3
    # Check for the number of input arguments
    if len(sys.argv) in [nargs_withoutpd, nargs_withpd]:
        # Check for steps to continue process
        try:
            steps = chckargs.checkStepsArg(sys.argv[1], config.proc_steps)
        except ValueError:
            print(proc_usage)
            sys.exit(0)
        if -1 not in steps:
            steps_comp = [False] * len(steps)
            # check for optional date input
            if len(sys.argv) == nargs_withoutpd:  # for operational running
                run_date = (
                    today - DT.timedelta(days=config.run_days_diff))
                run_processes = True
            elif len(sys.argv) == nargs_withpd:  # for manual running
                year = int(sys.argv[2])
                month = int(sys.argv[3])
                day = int(sys.argv[4])
                run_date = DT.datetime(year, month, day)
                run_processes = True
            else:
                print(proc_usage)
                sys.exit(0)
        else:
            print(proc_usage)
            sys.exit(0)
    return run_date, today, run_processes, steps, steps_comp, proc_usage


class run(object):

    # Dataset description
    DS_TYPE = 'WaterPoint'
    # Dataset periodicity
    DS_PERIODICITY = 'Daily'

    SCALED_DIR_NAME = 'scaled'

    def __init__(self, run_date, today, config):

        self.run_date = run_date

        self.prj_transformation = config.prj_transformation

        self.modelConfig = config.modelConfig

        self.mtabb_date = self.run_date.strftime("%b %d, %Y")

        self.email_subject = (
            config.dataset_region + " " + self.DS_PERIODICITY + " " +
            self.DS_TYPE + " for " +
            calendar.month_abbr[self.run_date.month] + " " +
            str(int(self.run_date.strftime("%d"))) +
            " " + str(self.run_date.year))

        log_dir = os.path.join(mconfig.base_path, constants.LOGS_DIR_NAME,
                               self.run_date.strftime("%Y%m%d"))

        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        log_file_name_prefix = (config.dataset_region + "-" +
                                self.DS_PERIODICITY + "-" +
                                self.DS_TYPE + "-" +
                                self.run_date.strftime("%Y%m%d") + "-" +
                                DT.datetime.now().strftime("%Y%m%d-%H%M%S"))

        self.log_path = os.path.join(log_dir, log_file_name_prefix +
                                     constants.LOG_EXT)

        # logging settings
        self.logger = print_msg.setLogger(self.log_path)

        wp_data_dir = os.path.join(mconfig.base_path,
                                   constants.DATA_DIR_NAME,
                                   config.dataset_region,
                                   self.DS_PERIODICITY, self.DS_TYPE)

        self.temp_workspace = os.path.join(wp_data_dir,
                                           constants.TEMP_DIR_NAME)
        temp_workspaces.setTempWorkspaces(self.temp_workspace)

        self.rain_ws = os.path.join(self.temp_workspace, 'RainGrid')

        self.rain_proj = os.path.join(self.rain_ws, 'rain_proj')

        self.rain_data_path = \
            os.path.join(config.rfe_rain_path, self.DS_PERIODICITY, 'RFE',
                         constants.GEOBILS_DIR_NAME,
                         self.run_date.strftime('%Y'))

        self.rfe_temp = os.path.join(self.temp_workspace, 'RFE')

        self.clet = os.path.join(wp_data_dir, 'clet')

        # Final depth paths
        final_depth = os.path.join(mconfig.remote_data_dir,
                                   config.dataset_region,
                                   self.DS_PERIODICITY,
                                   self.DS_TYPE,
                                   'final_depth')

        self.final_depthDB = os.path.join(final_depth, 'DailySampleData')

        self.final_depthGrid = os.path.join(final_depth, 'DailySampleGrid')

        self.txt_files = os.path.join(wp_data_dir, 'txt')

        if not os.path.exists(self.txt_files):
            os.makedirs(self.txt_files)

        # final_depth date
        self.day_before_run_date = self.run_date - DT.timedelta(days=1)

        self.wp_templates_dir = os.path.join(mconfig.templates_path,
                                             self.DS_TYPE)

        self.prj_file = os.path.join(self.wp_templates_dir,
                                     config.prj_file_name)

        self.shareOutputDir = config.manohar_output_dir

        notification_email_list_csv = os.path.join(
            mconfig.base_path,
            mconfig.email_list_directory_name,
            mconfig.notification_email_list_csv)

        self.notification_email_list = email_list.get_email_list(
            notification_email_list_csv, config.dataset_region,
            self.DS_PERIODICITY, self.DS_TYPE.lower())

        # web txt distribution dictionary
        self.web_txt_distribution = config.web_txt_distribution

        self.untracked_waterpoints = config.untracked_waterpoints

        self.web_ops_paths = mconfig.web_ops_paths

        self.ppg_doc = "TBD"
        # FIXME - sys,executable not getting the conda python path
        # when running from scheduler task, instead using windows python
        # system path.
        print_msg.print_process_info(sys.executable, sys.argv,
                                     self.ppg_doc, self.log_path)

        # Set geographic transformation
        gt.create_custom_gtf(constants.WGS84_SP_NAME, self.prj_file,
                             self.prj_transformation, arcpy, self.logger)

    def createRainGrid(self):
        """Create rain grid for waterhole site sampling."""
        t1 = DT.datetime.now()
        msg = ('Create Projected Rain Grid for {}'.format(self.mtabb_date))
        print_msg.add_process_step_header(t1, msg, self.logger)
        step_comp = False

        try:
            if not os.path.exists(self.rain_ws):
                os.makedirs(self.rain_ws)

            if not os.path.exists(self.rfe_temp):
                os.makedirs(self.rfe_temp)

            year = self.run_date.strftime('%Y')
            julian = dates.julian_day(self.run_date)

            rain_zip_name = 'rain_' + str(year) + str(julian) + '.tar.gz'
            rain_file_path = os.path.join(self.rain_data_path, rain_zip_name)

            if os.path.exists(rain_file_path):

                self.logger.info("Input: {}".format(rain_file_path))

                zips.unzipTar(rain_file_path, self.rfe_temp)

                arcpy.env.workspace = self.rain_ws

                rain_file_name = 'rain_' + str(year) + str(julian) + '.bil'

                rain_bil = os.path.join(self.rfe_temp, rain_file_name)

                rain_grid = os.path.basename(rain_bil[:-4])
                arcpy.CopyRaster_management(rain_bil, rain_grid)

                arcpy.ProjectRaster_management(
                    rain_grid, self.rain_proj, self.prj_file,
                    geographic_transform=self.prj_transformation)

                spatial_ref = arcpy.Describe(
                    self.rain_proj).spatialReference.name
                if os.path.exists(self.rain_proj):
                    msg = ('Output: {} projected to {} -> {}'.format(
                        rain_grid, spatial_ref, self.rain_proj))
                    self.logger.info(msg)
                    step_comp = True
            else:
                raise IOError("{} input does not exists".format(
                    rain_file_path))

            t2 = DT.datetime.now()
            print_msg.add_process_step_footer(t1, t2, step_comp, self.logger)
        except BaseException:
            msg = print_msg.get_traceback_msg(sys.exc_info())
            self.logger.error(msg)
            msg = "INCOMPLETE"
            self.logger.error(msg)
        return step_comp

    def model(self):
        '''
        Description: Runs waterpoint model for different regions defined by
                         basins300 GRID
        TODO:
          - check with Manohar how can be improve the GIS analysis process
            since DailySampleGrid is consuming lots of space.
            CY talked to Manohar and we could do this change in the future.
        '''
        t1 = DT.datetime.now()
        msg = ("Run Water Point model by region extent(s) for {}".format(
            self.mtabb_date))
        print_msg.add_process_step_header(t1, msg, self.logger)
        step_comp = False

        try:

            arcpy.env.workspace = self.temp_workspace

            # Set input data
            # Rainfall
            if os.path.exists(self.rain_proj):
                msg = 'Input: Rainfall -> {}'.format(self.rain_proj)
                self.logger.info(msg)
            else:
                raise IOError("{} input does NOT exists".format(
                    self.rain_proj))

            # Climatolgical ET
            clet = os.path.join(self.clet, 'clet_18{:02d}{:02d}'.format(
                self.run_date.month, self.run_date.day))
            if os.path.exists(clet):
                msg = 'Input: CLET -> {}'.format(clet)
                self.logger.info(msg)
            else:
                raise IOError("{} input does NOT exists".format(clet))
            self.logger.info(constants.SEP2)

            # Scale CLET from mm to m
            scaled_clet = arcpy.sa.Raster(clet) * 0.001
            scaled_clet.save('scaled_clet')

            # Scale projected rain from mm to m
            scaled_rain = arcpy.sa.Raster(self.rain_proj) * 0.001
            scaled_rain.save('scaled_rain')

            for region, inputs in self.modelConfig.iteritems():

                region_workspace = os.path.join(self.temp_workspace, region)
                if not os.path.exists(region_workspace):
                    os.makedirs(region_workspace)

                temp_workspaces.cleanTempWorkspaces(region_workspace)

                msg = '{}'.format(region)
                self.logger.info(msg)

                final_depth_grid_prefix = inputs['final_depth']
                templates_path = os.path.join(self.wp_templates_dir, region)
                wsh_area = os.path.join(templates_path,
                                        inputs['wsh_area']['grid'])
                wp_area = os.path.join(templates_path,
                                       inputs['wp_area']['grid'])
                sample = os.path.join(templates_path,
                                      inputs['sample']['grid'])
                basins = os.path.join(templates_path,
                                      inputs['basin']['grid'])
                wsh_scale_factor = inputs['wsh_area']['scale_factor']
                wp_scale_factor = inputs['wp_area']['scale_factor']

                # Check for region inputs
                for indata in [basins, wp_area, wsh_area, sample]:
                    if os.path.exists(indata):
                        msg = 'Input: {}'.format(indata)
                        self.logger.info(msg)
                    else:
                        raise IOError("{} input does NOT exists".format(
                            indata))
                # Set other region inputs

                # Sets the region workspace
                arcpy.env.workspace = region_workspace

                # Sets arcpy enviroment to subregion extent and cell size
                arcpy.env.cellSize = arcpy.sa.Raster(basins).meanCellWidth
                arcpy.env.extent = arcpy.sa.Raster(basins).extent

                zrain = arcpy.sa.ZonalStatistics(
                    basins, 'VALUE', os.path.join(
                        self.temp_workspace, 'scaled_rain'), 'MEAN', "DATA")
                zrain.save('zrain')  # do NOT delete, used in formatOutput()

                zevap = arcpy.sa.ZonalStatistics(
                    basins, 'VALUE', os.path.join(
                        self.temp_workspace, 'scaled_clet'), 'MEAN', "DATA")
                zevap.save('zevap')  # do NOT delete, used in formatOutput()

                volume = ((0.05 * zrain) * (arcpy.sa.Raster(wsh_area) *
                                            wsh_scale_factor))

                depth_v = volume / (arcpy.sa.Raster(wp_area) * wp_scale_factor)

                ninety_days_before_run_date = self.day_before_run_date - \
                    DT.timedelta(days=90)

                previous_final_depth_grid_path = os.path.join(
                    self.final_depthGrid, region,
                    final_depth_grid_prefix +
                    self.day_before_run_date.strftime('%y%m%d'))

                ninety_day_before_previous_final_depth_grid_path = \
                    os.path.join(
                        self.final_depthGrid, region,
                        final_depth_grid_prefix +
                        ninety_days_before_run_date.strftime('%y%m%d'))

                if arcpy.Exists(previous_final_depth_grid_path):
                    msg = "Input: Prev final depth: {}".format(
                        previous_final_depth_grid_path)
                    self.logger.info(msg)
                    water_level = (
                        arcpy.sa.Raster(previous_final_depth_grid_path) +
                        depth_v -
                        zevap - 0.003)
                # The first day's of the grid's calculation
                elif self.run_date.strftime('%Y%m%d') == 20010101:
                    water_level = (depth_v - zevap - 0.003)
                else:
                    raise IOError("Incorrect input date")

                # Delete the ninety_day_before_previous_final_depth_grid_path
                if arcpy.Exists(
                        ninety_day_before_previous_final_depth_grid_path):
                    arcpy.Delete_management(
                        ninety_day_before_previous_final_depth_grid_path)

                dptmp = arcpy.sa.Con(water_level < 0, 0, water_level)

                final_depth_grid_name = (final_depth_grid_prefix +
                                         self.run_date.strftime('%y%m%d'))
                f_depth = arcpy.sa.Con(dptmp > 2, 2, dptmp)
                f_depth.save(final_depth_grid_name)

                del water_level, depth_v

                final_depth_grid_path = os.path.join(self.final_depthGrid,
                                                     region)
                if not os.path.exists(final_depth_grid_path):
                    os.makedirs(final_depth_grid_path)
                remote_final_depth_grid_path = os.path.join(
                    final_depth_grid_path, final_depth_grid_name)
                local_final_depth_grid_path = os.path.join(
                    region_workspace, final_depth_grid_name)

                arcpy.CopyRaster_management(
                    local_final_depth_grid_path,
                    remote_final_depth_grid_path)

                msg = 'Output: Final depth -> {}'.format(
                    remote_final_depth_grid_path)
                self.logger.info(msg)

                sample_list = [zrain, zevap, volume, f_depth]

                # TODO - the DBF file and path could be changed to use folder
                # name as DBF file, needs to be done from model()
                # Example:
                # ..\Africa\Daily\WaterPoint\final_depth\DailySampleData
                # \Fewsnet\finalDepthTable_190215.dbf
                out_dir = os.path.join(
                    self.final_depthDB,
                    region,
                    self.run_date.strftime('finalDepthTable_%y%m%d'))

                if not os.path.exists(out_dir):
                    os.makedirs(out_dir)

                output_table = os.path.join(out_dir,
                                            self.run_date.strftime(
                                                'finalDepthTable_%y%m%d.dbf'))

                arcpy.env.cellSize = arcpy.sa.Raster(sample).meanCellWidth
                arcpy.env.extent = arcpy.sa.Raster(sample).extent

                arcpy.gp.Sample_sa(sample_list, sample, output_table,
                                   "NEAREST", "VALUE", "CURRENT_SLICE")

                msg = 'Output: Sample table -> {}'.format(output_table)
                self.logger.info(msg)

                share_daily = os.path.join(self.shareOutputDir, region,
                                           os.path.basename(out_dir))
                if not os.path.exists(share_daily):
                    os.makedirs(share_daily)

                copy_res = file_utilities.copy_file(output_table, share_daily)
                if copy_res == 1:
                    msg = 'Output for sharing: {}'.format(share_daily)
                    self.logger.info(msg)
                else:
                    raise IOError("{} copy FAILED".format(share_daily))

                if (os.path.exists(output_table) and
                        os.path.exists(share_daily)):
                    step_comp = True
                temp_workspaces.cleanTempWorkspaces(region_workspace)
                self.logger.info(constants.SEP2)

            t2 = DT.datetime.now()
            print_msg.add_process_step_footer(t1, t2, step_comp, self.logger)
        except BaseException:
            msg = print_msg.get_traceback_msg(sys.exc_info())
            self.logger.error(msg)
            msg = "INCOMPLETE"
            self.logger.error(msg)
        finally:
            temp_workspaces.cleanTempWorkspaces(self.temp_workspace)
        return step_comp

    def formatOutput(self):
        """Format output to create CSV TXT time series files for each site."""
        t1 = DT.datetime.now()

        msg = ('Format model outputs to TXT files for {}'.format(
            self.mtabb_date))
        print_msg.add_process_step_header(t1, msg, self.logger)
        step_comp = False

        try:
            # Set remote data sharing folder
            rem_scaled_folder = os.path.join(
                self.shareOutputDir, self.SCALED_DIR_NAME)
            if not os.path.exists(rem_scaled_folder):
                os.makedirs(rem_scaled_folder)

            msg = "Output: Local TXT files location -> {}".format(
                self.txt_files)
            self.logger.info(msg)
            self.logger.info(constants.SEP2)

            # To count all the TXT files processed
            text_files_count = 1

            for region, inputs in self.modelConfig.iteritems():

                # To count for all the TXT files processed by region
                text_files_per_region = 0

                msg = '{}'.format(region)
                self.logger.info(msg)

                # Set table path and raise exception if not found.
                table = os.path.join(
                    self.final_depthDB, region,
                    self.run_date.strftime('finalDepthTable_%y%m%d'),
                    self.run_date.strftime('finalDepthTable_%y%m%d.dbf'))

                if not os.path.exists(table):
                    msg = ('Final depth outputs tables for {} do NOT exists, '
                           'run model process step for this '
                           'date\n'.format(self.run_date.strftime('%y%m%d')))
                    self.logger.error(msg)
                    raise Exception

                run_date_ymd_format = self.run_date.strftime('%Y%m%d')

                # Read DBF table and get columns using pandas dataframe
                dbf = DBF(table)
                frame = pd.DataFrame(iter(dbf))
                col_headers = list(frame)

                # Get final depth column name
                dep = next(i for i in col_headers
                           if i.startswith(inputs['final_depth']))

                # Get sample ID column name
                index = next(i for i in col_headers
                             if inputs['sample']['grid'].startswith(i)
                             )

                for water_hole, row in frame.iterrows():

                    if int(row[index]) in self.untracked_waterpoints:
                        continue

                    # Set TXT filename
                    fName = ('scaled_{}{}'.format(int(row[index]),
                                                  constants.TXT_EXT))

                    # Local TXT file output
                    in_file = os.path.join(self.txt_files, fName)
                    self.logger.debug(in_file)

                    # Load the text file time series for a location into
                    # a data frame
                    df = pd.read_csv(in_file, sep=" ", dtype=object)

                    df = df.astype({'Date': 'int32',
                                    'Rain(m)': 'float',
                                    'Evap(m)': 'float',
                                    'Depth(m)': 'float',
                                    'Scaled_Depth(%)': 'float'})

                    # The existing scaled_###.txt file might be appended with
                    # new data or updated starting at the specified date.

                    # Retain records earlier than current run date.
                    run_date_int = int(self.run_date.date().strftime('%Y%m%d'))
                    df = df[df.Date < run_date_int]

                    # Make sure the current run date is one day later than
                    # the last retainded record.
                    last_record_date = DT.datetime.strptime(
                         str(df.Date.iloc[-1]), '%Y%m%d')

                    if (self.run_date - last_record_date).days != 1:
                        msg = "Time series not continuous %s minus %s is not 1"
                        raise Exception(msg % (self.run_date,
                                               last_record_date))

                    # Determine historical maximum depth including
                    # current run date.
                    maxdep = df['Depth(m)'].append(pd.Series(row[dep])).max()

                    sdep = (row[dep] / maxdep) * 100
                    rain = row['zrain']
                    evap = row['zevap']
                    depth = row[dep]

                    # Append data for the current run date to time series.
                    modDf = df.append(pd.Series([run_date_ymd_format,
                                                str(rain), str(evap),
                                                str(depth), str(sdep)],
                                      index=df.columns),
                                      ignore_index=True).fillna(0)

                    if modDf is not None:
                        modDf.to_csv(in_file, sep=" ", index=False)
                        if os.path.exists(in_file):
                            msg = 'Output: {} created'.format(os.path.basename(
                                in_file))
                            self.logger.info(msg)
                            # Copy local TXT output to data sharing location
                            if mconfig.transfer_files:

                                out_txt = os.path.join(rem_scaled_folder,
                                                       fName)
                                copy_res = file_utilities.copy_file(in_file,
                                                                    out_txt)
                                if copy_res == 1:
                                    text_files_count += 1
                            text_files_per_region += 1
                        else:
                            raise IOError("{} output does NOT exists".format(
                                os.path.basename(in_file)))

                self.logger.info("{}s {} processed for this region".format(
                    constants.TXT_EXT[1:].upper(), text_files_per_region))
                self.logger.info(constants.SEP2)

            if mconfig.transfer_files:
                # Get number of local and remote TXT files created
                out_txt_n = glob.glob(os.path.join(
                    self.txt_files, '*' + constants.TXT_EXT))
                rem_out_txt_n = glob.glob(os.path.join(
                    rem_scaled_folder, '*' + constants.TXT_EXT))
                msg = ("Output: {} files: {} created, {} shared, "
                       "{} excpected".format(constants.TXT_EXT[1:].upper(),
                                             text_files_count,
                                             len(rem_out_txt_n),
                                             len(out_txt_n)))

                if text_files_count == len(out_txt_n) and \
                   text_files_count == len(rem_out_txt_n):
                    self.logger.info(msg)
                else:
                    raise IOError(msg)

                self.logger.info(constants.SEP2)

            # Use scaled TXT files to copy files on web by collaborator(s)
            # using self.web_txt_distribution dict info

                count = 0

                for web_txt in self.web_txt_distribution.keys():
                    self.logger.info("Web sharing for {}".format(web_txt))
                    wp_txt_list = \
                        self.web_txt_distribution[web_txt]['wp_txt_points']

                    for wp_txt_n in wp_txt_list:
                        scaled_txt_file = os.path.join(
                            self.txt_files,
                            'scaled_{}{}'.format(wp_txt_n, constants.TXT_EXT))
                        if os.path.exists(scaled_txt_file):

                            for web_path in self.web_ops_paths:
                                web_dir_path = os.path.join(
                                    web_path,
                                    self.DS_TYPE.lower().replace("point",
                                                                 "_point"),
                                    web_txt)
                                if not os.path.exists(web_dir_path):
                                    os.makedirs(web_dir_path)
                                dst_file_path = os.path.join(
                                    web_dir_path,
                                    os.path.basename(scaled_txt_file))
                                file_transfer = file_utilities.copy_file(
                                    scaled_txt_file, dst_file_path)
                                if file_transfer == 1:
                                    msg = ("{} copied to {}".format(
                                        os.path.basename(scaled_txt_file),
                                        web_dir_path))
                                    self.logger.info(msg)
                                    if self.web_ops_paths.index(web_path) == 0:
                                        count += 1
                                else:
                                    raise IOError(
                                        "{} NOT copied to web at {}".format(
                                            os.path.basename(scaled_txt_file),
                                            web_dir_path))
                        else:
                            raise IOError("{} does NOT exits".format(
                                scaled_txt_file))
                    msg = ("Output for Web: {} copied, {} expected".format(
                        count, len(wp_txt_list)))
                    if count == len(wp_txt_list):
                        self.logger.info(msg)
                        step_comp = True
                    else:
                        raise IOError(msg)
            else:
                self.logger.info("Transfer files disable")
                step_comp = True

            t2 = DT.datetime.now()
            print_msg.add_process_step_footer(t1, t2, step_comp, self.logger)
        except BaseException:
            msg = print_msg.get_traceback_msg(sys.exc_info())
            self.logger.error(msg)
            msg = "INCOMPLETE"
            self.logger.error(msg)
        return step_comp

    @staticmethod
    def __getConditionColor(test_median, median):
        colors = ['00d000', 'fff000', 'd09000', 'e03030', '777777']
        if test_median > median:
            return colors[0]
        elif (test_median / median) > .5:
            return colors[1]
        elif (test_median / median) > .03:
            return colors[2]
        elif (test_median == 0 and median == 0):
            return colors[4]
        else:
            return colors[3]

    def updateDataBase(self):
        """Update database with new copies of the time series data."""
        t1 = DT.datetime.now()
        msg = ("Update Web DB with {} files information for {}".format(
            constants.TXT_EXT[1:].upper(), self.mtabb_date))
        print_msg.add_process_step_header(t1, msg, self.logger)
        step_comp = False

        try:
            path_2_files = []

            if os.path.exists(self.txt_files):
                if os.listdir(self.txt_files):
                    files = os.listdir(self.txt_files)
                    for file_ in files:
                        path_2_files.append(os.path.join(self.txt_files,
                                                         file_))
            else:
                raise IOError("{} does NOT exists".format(self.txt_files))

            run_date_ymd_format = self.run_date.strftime('%Y%m%d')

            # select one of the text files
            in_file = path_2_files[0]

            # Load the text file in to a data frame
            df = pd.read_csv(in_file, sep=" ", dtype=object)

            date_col = df['Date']

            # convert the date_col pandas object in to a list
            date_list = date_col.values.tolist()

            # Check if the latest date in text file is same as run date
            if date_list[-1] == run_date_ymd_format:

                conn_res0, conn_res1, conn_res2, conn_res3, conn_res4 = \
                    False, False, False, False, False
                msg = 'Running updates on "median_depth table" '
                self.logger.info(msg)
                msg = 'TRUNCATING "median_depth table" '
                self.logger.info(msg)
                remove_median_depth_data = 'TRUNCATE median_depth RESTART ' \
                                           'IDENTITY'
                conn_res0 = cursor.exec_on_pg(remove_median_depth_data,
                                              self.logger)
                if not conn_res0:
                    raise Exception("{} ... FAILED".format(
                        remove_median_depth_data))

                msg = 'Running updates on "location data table" '
                self.logger.info(msg)
                msg = 'TRUNCATING "location data table" '
                self.logger.info(msg)
                remove_location_data = 'TRUNCATE location_data RESTART ' \
                                       'IDENTITY'
                conn_res1 = cursor.exec_on_pg(remove_location_data,
                                              self.logger)
                if not conn_res1:
                    raise Exception("{} ... FAILED".format(
                        remove_location_data))
                msg = 'Running updates on "locations table" '
                self.logger.info(msg)
                msg = 'Running web viewer updates'
                self.logger.info(msg)

                for file_ in path_2_files:
                    with open(file_, 'r') as current_file:
                        this_location_id = \
                            os.path.basename(current_file.name).split(
                                os.sep).pop().split('.')[0].split(
                                '_').pop()
                        depth_list_and_median = {}
                        lines = current_file.readlines()
                        location_data_sql = ''
                        for line in lines:
                            if line == lines[0]:
                                columns = line.lower().strip()
                                columns = re.sub('\s', ',', columns)
                                columns = re.sub('\\(.+?\)', '', columns)
                                location_data_sql += (
                                    'insert into location_data (' +
                                    columns + ', location_id,  '
                                    'given_date, "day") '
                                    'values ')
                            else:
                                column_data = line.lower().strip().split()

                                date = column_data[0]
                                del (column_data[0])

                                year = date[:4]
                                month = date[4:6]
                                day = date[6:]

                                days = str(month) + str(day)
                                days = int(days)

                                date_ymd_format = DT.datetime(
                                    int(year), int(month), int(day)).strftime(
                                    '%Y-%m-%d')

                                depth = float(column_data[3])

                                if str(days) not in depth_list_and_median:
                                    depth_list_and_median[str(days)] = {
                                        'total': [],
                                        'median': None}

                                depth_list_and_median[
                                    str(days)]['total'].append(depth)

                                column_data[0] = float(column_data[0]) * 1000
                                column_data[1] = float(column_data[1]) * 1000

                                col_data =\
                                    ','.join(str(value) for value in
                                             column_data)

                                if line == lines[-1]:
                                    location_data_sql += (
                                        '(\'' + date_ymd_format +
                                        '\'' + ',' + col_data + ',' +
                                        str(this_location_id) + ',' +
                                        date + ',' +
                                        str(days) + ')')
                                else:
                                    location_data_sql += (
                                        '(\'' + date_ymd_format +
                                        '\'' + ',' + col_data + ',' +
                                        str(this_location_id) + ',' +
                                        date + ',' +
                                        str(days) + '),')

                        conn_res2 = cursor.exec_on_pg(location_data_sql,
                                                      self.logger)
                        if not conn_res2:
                            raise Exception("{} ... FAILED".format(
                                location_data_sql))

                        median_depth_sql = (
                            'insert into median_depth ' +
                            '(days, "medianDepth", location_id) values ')

                        # Calculate median depth of each day for all the years
                        for dayKey in depth_list_and_median:
                            depth_list = np.array(
                                depth_list_and_median[dayKey]['total'])

                            depth_list_and_median[dayKey]['median'] = \
                                np.median(depth_list)

                            median_depth_sql += ('(' +
                                                 str(dayKey) +
                                                 ', ' +
                                                 str(depth_list_and_median[
                                                     dayKey]['median']) +
                                                 ', ' +
                                                 str(this_location_id) +
                                                 '),')

                        conn_res3 = cursor.exec_on_pg(median_depth_sql[:-1],
                                                      self.logger)
                        if not conn_res3:
                            raise Exception("{} ... FAILED".format(
                                median_depth_sql[:-1]))

                        # Get the color of run date
                        color = run.__getConditionColor(
                            depth_list_and_median[str(days)]['total'][-1],
                            depth_list_and_median[str(days)]['median'])

                        location_sql = (
                            'update locations set "graphStyle" = ' +
                            '\'' + color + '\'' + ' where uid = ' +
                            this_location_id)

                        conn_res4 = cursor.exec_on_pg(location_sql,
                                                      self.logger)
                        if not conn_res4:
                            raise Exception("{} ... FAILED".format(
                                median_depth_sql[:-1]))

                if (conn_res0 and conn_res1 and conn_res2 and conn_res3 and
                        conn_res4):
                    self.logger.info("All Web DB updates completed")
                    step_comp = True
            else:
                msg = (
                    "{} latest date data found in the text files. "
                    "Check {} file {}\n".format(date_list[-1],
                                                constants.TXT_EXT[1:].upper(),
                                                in_file))
                self.logger.error(msg)
                raise Exception

            t2 = DT.datetime.now()
            print_msg.add_process_step_footer(t1, t2, step_comp, self.logger)
        except BaseException:
            msg = print_msg.get_traceback_msg(sys.exc_info())
            self.logger.error(msg)
            msg = "INCOMPLETE"
            self.logger.error(msg)
        return step_comp

    def sendEmail(self):
        """Send email notification."""
        email_file.send(self.email_subject, self.notification_email_list,
                        self.log_path)
