# coding=utf8

"""
@author: Bruce Worstell
Created August 16, 2021

Description: Script to support fixing the Waterpoint Viewer scaled_###.txt
    files. This code analyzes the depth column of the time series data for
    each point and computes a running max depth for each record based on
    depths for previous dates and calculates new values for Scaled_Depth(%).
"""

import os
import sys
import argparse
from glob import glob
import pandas as pd


def load_scaled_txt(text_fp):
    """Load a scaled_###.txt file as a Pandas DataFrame."""
    scaled_txt_df = pd.read_csv(text_fp, sep=' ', dtype=object)

    # Fix column types
    scaled_txt_df = scaled_txt_df.astype({'Date': 'int32',
                                          'Rain(m)': 'float',
                                          'Evap(m)': 'float',
                                          'Depth(m)': 'float',
                                          'Scaled_Depth(%)': 'float'})

    return scaled_txt_df


def calc_scaled_depth(df):
    """
    Re-compute the Scaled_Depth(%).

    For each date record, re-compute scaled depth based on prior history.
    """
    # Calculate running max of Depth(m) based on all previous records
    # in the time series.
    df['cum_max_dep'] = df['Depth(m)'].cummax()

    # Old logic comments below for documentation.
    # Apply logic to set maxdep used to compute Scaled_Depth(%).

    # If the cumulative/running max depth is less than 1,
    # maxdep is set to 1

    # df.loc[df['cum_max_dep'] < 1, 'maxdep']  = 1

    # If the cumulative/running max depth is greater than 1,
    # maxdep is set to the cumulative/running max depth

    # df.loc[df['cum_max_dep'] >= 1, 'maxdep']  = df['cum_max_dep']

    # Note. Maximum detph possible is 2 which is set from the
    # source raster data.

    # Compute scaled depth based on historical maximum depth
    df['Scaled_Depth(%)'] = 100 * df['Depth(m)'] / df['cum_max_dep']

    return df[['Date', 'Rain(m)', 'Evap(m)', 'Depth(m)',
               'Scaled_Depth(%)']].fillna(0)


def get_args(argv=None):
    """Process commandline arguments."""
    # Command line with flags using argparse
    parser = argparse.ArgumentParser(description='Recomputes the Scaled_Depth '
                                     'for each of the scaled_###.txt files '
                                     'using the running historical max depth.')

    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='verbose flag')

    parser.add_argument('input_path',
                        help=('Input path to the existing scaled text files.'))

    parser.add_argument('output_path',
                        help=('Non-existing output path where the new files '
                              'will be created.'))

    return parser.parse_args(argv)


if __name__ == '__main__':

    args = get_args(sys.argv[1:])

    # Get list of input scaled files.
    txt_files = glob(os.path.join(args.input_path, 'scaled_*.txt'))
    print('Processing %s files' % len(txt_files))

    if not txt_files:
        raise Exception('No scaled_######.txt files at %s' % args.input_path)

    if os.path.exists(args.output_path):
        raise Exception('Output Path "%s" already exists.' % args.output_path)
    else:
        # Create new target path.
        os.makedirs(args.output_path)

    for txt_fp in txt_files:

        # Load scaled_#####.txt file into Pandas DataFrame.
        df = load_scaled_txt(txt_fp)

        # Re-calculate Scaled_Depth(%) using historical data records.
        mod_df = calc_scaled_depth(df)

        output_scaled_fp = os.path.join(args.output_path,
                                        os.path.basename(txt_fp))

        # Write a new scaled text file with new Scaled_Depth(%) values.
        mod_df.to_csv(output_scaled_fp, sep=' ', index=False)

        if args.verbose:
            print('Creating output %s' % output_scaled_fp)

    print('New files created in %s' % os.path.dirname(output_scaled_fp))
